-- booleans

data Bool : Set
{ true  : Bool
; false : Bool
}

fun and : Bool -> Bool -> Bool
{ and true  x = x
; and false x = false
}

-- identity and logic

data Id (A : Set) (a : A) : A -> Set
{ refl : Id A a a
}

fun cong : (A : Set) -> (a : A) -> (a' : A) -> 
           (B : Set) -> (f : A -> B) -> Id A a a' -> Id B (f a) (f a')
{ cong A a .a B f (refl .A .a) = refl B (f a)
}

fun True : Bool -> Set
{ True b = Id Bool b true
}
let triv : True true
         = refl Bool true

fun False : Bool -> Set
{ False b = Id Bool b false
}
let triv' : False false
          = refl Bool false

-- lists

data List (A : Set) : Set 
{ nil  : List A
; cons : A -> List A -> List A
}

fun all : (A : Set) -> (A -> Bool) -> List A -> Bool
{ all .A p (nil A) = true
; all .A p (cons A a as) = and (p a) (all A p as)
}

fun filter : (A : Set) -> (A -> Bool) -> List A -> List A
{ filter .A p (nil A) = nil A
; filter .A p (cons A a as) = case p a
  { true  -> cons A a (filter A p as)
  ; false -> filter A p as
  }
} 

fun lemAllFilter : (A : Set) -> (p : A -> Bool) -> (l : List A) ->
  True (all A p (filter A p l))
{ lemAllFilter .A p (nil A) = triv
; lemAllFilter .A p (cons A a as) = case p a
  { true  -> lemAllFilter A p as
  ; false -> lemAllFilter A p as
  }
} 

fun map : (A : Set) -> (B : Set) -> (A -> B) -> List A -> List B
{ map A B f (nil .A) = nil B
; map A B f (cons .A a as) = cons B (f a) (map A B f as)
}

fun lemMapFilter : (A : Set) -> (B : Set) -> (f : A -> B) -> (p : B -> Bool) ->
  (l : List A) ->
  Id (List B) (filter B p (map A B f l)) 
              (map A B f (filter A (\ x -> p (f x)) l)) 
{ lemMapFilter A B f p (nil .A)       = refl (List B) (nil B)
; lemMapFilter A B f p (cons .A a as) = case p (f a)
  { false -> lemMapFilter A B f p as
  ; true  -> cong (List B) (filter B p (map A B f as)) 
                           (map A B f (filter A (\ x -> p (f x)) as)) 
                  (List B) (\ bs -> cons B (f a) bs) 
                  (lemMapFilter A B f p as)
  }
} 

data Nat : Set
{ zero : Nat
; suc  : Nat -> Nat
}

fun take : (A : Set) -> Nat -> List A -> List A
{ take A zero    xs             = nil A
; take A (suc n) (nil .A)       = nil A
; take A (suc n) (cons .A x xs) = cons A x (take A n xs)
}

fun drop : (A : Set) -> Nat -> List A -> List A
{ drop A zero    xs             = xs
; drop A (suc n) (nil .A)       = nil A
; drop A (suc n) (cons .A x xs) = drop A n xs
}

data Prod (A : Set) (B : Set) : Set
{ pair : A -> B -> Prod A B
}

fun splitAt : (A : Set) -> Nat -> List A -> Prod (List A) (List A)
{ splitAt A zero    xs             = pair (List A) (List A) (nil A) xs
; splitAt A (suc n) (nil .A)       = pair (List A) (List A) (nil A) (nil A)
; splitAt A (suc n) (cons .A x xs) = case splitAt A n xs
  { (pair .(List A) .(List A) ys zs) -> pair (List A) (List A) (cons A x ys) zs
  }
}

fun splitAtDef : (A : Set) -> (n : Nat) -> (l : List A) -> 
  Id (Prod (List A) (List A)) 
     (splitAt A n l) 
     (pair (List A) (List A) (take A n l) (drop A n l))
{ splitAtDef A zero xs = 
   refl (Prod (List A) (List A)) 
        (pair (List A) (List A) (nil A) xs)
; splitAtDef A (suc n) (nil .A) = 
   refl (Prod (List A) (List A)) 
        (pair (List A) (List A) (nil A) (nil A))
; splitAtDef A (suc n) (cons .A x xs) = case splitAt A n xs
  { (pair .(List A) .(List A) ys zs) ->
      cong (Prod (List A) (List A)) 
             (splitAt A n xs)
             (pair (List A) (List A) (take A n xs) (drop A n xs)) 
           (Prod (List A) (List A)) 
             (\ p -> case p 
                { (pair .(List A) .(List A) ys zs) -> 
                     pair (List A) (List A) (cons A x ys) zs })
           (splitAtDef A n xs)
  }
}